// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef NCMPC_OUTPUTS_PAGE_HXX
#define NCMPC_OUTPUTS_PAGE_HXX

#include "config.h"

#ifdef ENABLE_OUTPUTS_SCREEN
struct PageMeta;
extern const PageMeta screen_outputs;
#endif /* ENABLE_OUTPUTS_SCREEN */

#endif
