// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#pragma once

#include <string_view>

void
InitXtermTitle() noexcept;

void
set_xterm_title(std::string_view title) noexcept;
