// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef HISTORY_HXX
#define HISTORY_HXX

#include <list>
#include <string>

using History = std::list<std::string>;

#endif
