// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef CUSTOM_COLORS_HXX
#define CUSTOM_COLORS_HXX

void
colors_define(short color, short r, short g, short b) noexcept;

void
ApplyCustomColors() noexcept;

#endif
