// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef NCMPC_SEARCH_PAGE_HXX
#define NCMPC_SEARCH_PAGE_HXX

#include "config.h"

#ifdef ENABLE_SEARCH_SCREEN
struct PageMeta;
extern const PageMeta screen_search;
#endif

#endif
