// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef NCMPC_HELP_PAGE_HXX
#define NCMPC_HELP_PAGE_HXX

#include "config.h"

#ifdef ENABLE_HELP_SCREEN
struct PageMeta;
extern const PageMeta screen_help;
#endif

#endif
