// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef NCMPC_CHAT_PAGE_HXX
#define NCMPC_CHAT_PAGE_HXX

#include "config.h"

#ifdef ENABLE_CHAT_SCREEN

struct PageMeta;
extern const PageMeta screen_chat;

#endif

#endif
