// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef NCMPC_QUEUE_PAGE_HXX
#define NCMPC_QUEUE_PAGE_HXX

struct PageMeta;

extern const PageMeta screen_queue;

#endif
