// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef NCMPC_KEY_DEF_PAGE_HXX
#define NCMPC_KEY_DEF_PAGE_HXX

#include "config.h"

#ifdef ENABLE_KEYDEF_SCREEN
struct PageMeta;
extern const PageMeta screen_keydef;
#endif /* ENABLE_KEYDEF_SCREEN */

#endif
